import numpy as np
from pydesy.basic import Point
from pydesy.gmath import *

class Ellipsoid:
    def __init__(self, a, b):
        self.a = a
        self.b = b
        self.e2 = (sqrt(self.a ** 2 - self.b ** 2) / self.a) ** 2

    def geodetic_to_rectangle(self, B, L, H):
        N = self.a / sqrt(1 - self.e2 * sin(B) ** 2)
        return {
            "X": (N + H) * cos(B) * cos(L),
            "Y": (N + H) * cos(B) * sin(L),
            "Z": ((1 - self.e2) * N + H) * sin(B)
        }

    def Bessel_inverse_geo_task(self, B1, L1, B2, L2):
        w1, w2 = sqrt(1 - self.e2 * sin(B1) ** 2), sqrt(1 - self.e2 * sin(B2) ** 2) #БВ
        sinu1, sinu2 = sin(B1) * sqrt(1 - self.e2) / w1, sin(B2) * sqrt(1 - self.e2) / w2 #БВ
        cosu1, cosu2 = cos(B1) / w1, cos(B2) / w2 #БВ
        l = L2 - L1 #Градусы
        a1, a2 = sinu1 * sinu2, cosu1 * cosu2 #БВ
        b1, b2 = cosu1 * sinu2, sinu1 * cosu2 #БВ

        A12 = 0
        preSigma = 0
        sigma, eps = 0, 0
        while abs(eps) > 1 / 10000 or eps != 0:
            lam = l + sigma
            p = cosu2 * sin(lam) #БВ
            q = B1 - B2 * cos(lam) #Градусы
            A12 = abs(atan(p / q)) #Градусы

            if p > 0 and q < 0: A12 = 180 - A12
            elif p < 0 and q < 0: A12 = 180 + A12
            elif p < 0 and q > 0: A12 = 360 - A12
            elif p == 0:
                if q > 0: A12 = 90
                else: A12 = 270
            elif q == 0:
                if p > 0: A12 = 0
                else: A12 = 180

            sind, cosd = p * sin(A12) + q * cos(A12), a1 + a2 * cos(lam)
            d = abs(atan(sind / cosd)) #Degrees

            if cosd < 0:
                d = 180 - d

            sinA0 = cosu1 * sin(A12)
            cos2A0 = 1 - sinA0 ** 2
            x = 2 * a1 - cos2A0 * cosd
            a = (e2 / 2 + e2 ** 2 / 8 + e2 ** 3 / 16) - (e2 ** 2 / 16 + e2 ** 3 / 16) * cos2A0 + 3 * e2 ** 3 / 128 * cos2A0 ** 2
            b = e2 ** 2  / 16 + e2 ** 3 / 16 - e2 ** 3 / 32 * cos2A0
            sigma = deg((a * rad(d) - b * x * sind) * sinA0)
            eps = sigma - preSigma
            preSigma = sigma

        print(A12)
        
class Helmert:
    def __init__(self, points1, points2):
        self.points1 = points1
        self.points2 = points2
        
    def calculate_params(self):
        assert len(self.points1) == len(self.points2) and len(self.points1) >= 3 and len(self.points2) >= 3
        matrixL = list()
        for i in range(len(self.points1)):
            matrixL.append([self.points2[i].x - self.points1[i].x])
            matrixL.append([self.points2[i].y - self.points1[i].y])
            matrixL.append([self.points2[i].z - self.points1[i].z])
        self.matrixL = np.array(matrixL)

        matrixA = list()
        for i in range(len(self.points1)):
            matrixA.append([1, 0, 0, self.points2[i].x, 0, self.points2[i].z, -self.points2[i].y])
            matrixA.append([0, 1, 0, self.points2[i].y, -self.points2[i].z, 0, self.points2[i].x])
            matrixA.append([0, 0, 1, self.points2[i].z, self.points2[i].y, -self.points2[i].x, 0])
        self.matrixA = np.array(matrixA)
        self.matrixX = -np.linalg.inv(np.transpose(self.matrixA) @ self.matrixA) @ np.transpose(self.matrixA) @ self.matrixL

        return {
            "x": self.matrixX[0,0],
            "y": self.matrixX[1,0],
            "z": self.matrixX[2,0],
            "m": 1 + self.matrixX[3,0],
            "ex\"": deg(self.matrixX[4,0]) * 3600,
            "ey\"": deg(self.matrixX[5,0]) * 3600,
            "ez\"": deg(self.matrixX[6,0]) * 3600
        }

    def calculate_accuracy_score(self):
        self.matrixV = self.matrixA @ self.matrixX + self.matrixL
        sigma0 = sqrt(np.sum(np.power(self.matrixV, 2)) / (3 * len(self.points1) - 7))
        self.matrixQ = np.linalg.inv(np.transpose(self.matrixA) @ self.matrixA)

        return {
            "σ0": sigma0,
            "σx" : sigma0 * sqrt(self.matrixQ[0, 0]),
            "σy" : sigma0 * sqrt(self.matrixQ[1, 1]),
            "σz" : sigma0 * sqrt(self.matrixQ[2, 2]),
            "σm" : sigma0 * sqrt(self.matrixQ[3, 3]),
            "σex" : deg(sigma0 * sqrt(self.matrixQ[4, 4])) * 3600,
            "σey" : deg(sigma0 * sqrt(self.matrixQ[5, 5])) * 3600,
            "σez" : deg(sigma0 * sqrt(self.matrixQ[6, 6])) * 3600
        }

    def calculate_divergences(self):
        divX, divY, divZ = list(), list(), list()

        for i in range(len(self.points1)):
            matrixB = np.array([
                [1, -self.matrixX[6, 0], self.matrixX[5, 0]],
                [self.matrixX[6, 0], 1, -self.matrixX[4, 0]],
                [-self.matrixX[5, 0], self.matrixX[4, 0], 1]
            ])

            move = Point(self.matrixX[0, 0], self.matrixX[1, 0], self.matrixX[2, 0]).getArrayV()
            controlPoint = (1 + self.matrixX[3, 0]) * matrixB @ self.points2[i].getArrayV() + move

            divX.append([controlPoint[0, 0] - self.points1[i].x])
            divY.append([controlPoint[1, 0] - self.points1[i].y])
            divZ.append([controlPoint[2, 0] - self.points1[i].z])

        return {
            "dx" : divX,
            "dy" : divY,
            "dz" : divZ
        }


import numpy as np
from pydesy.gmath import *

class Point:
    def __init__(self, x=0, y=0, z=0):
        self.x = x
        self.y = y
        self.z = z
        
    def get_dict(self):
        return {"x" : self.x, "y" : self.y, "h" : self.z}

    def get_list_h(self):
        return [self.x, self.y, self.z]

    def get_list_v(self):
        return [
            [self.x],
            [self.y],
            [self.z]
        ]

    def get_array_h(self):
        return np.array(self.getList())

    def get_array_v(self):
        return np.array([
            [self.x],
            [self.y],
            [self.z]
        ])

    def dgt(self, d=0, a=0):
        return direct_geo_task(self.x, self.y, d, a)
    
    def igt(self, point):
        return inverse_geo_task(self.x, self.y, point.x, point.y)


    def reverse(self):
        self.x, self.y = self.y, self.x
        
    def __add__(self, other):
        if isinstance(other, Point):
            return Point(other.x + self.x, other.y + self.y, other.z + self.z)
        else:
            return NotImplemented
    
    def __mul__(self, other):
        if isinstance(other, Point):
            return Point(other.x * self.x, other.y * self.y, other.z * self.z)
        
        elif isinstance(other, int) or isinstance(other, float):
            return Point(other * self.x, other * self.y, other * self.z)

        else:
            return NotImplemented
        
    def __neg__(self):
        return Point(-self.x, -self.y, -self.z)

    def __pos__(self):
        return Point(self.x, self.y, self.z)
    
    def __repr__(self):
        return f"x: {self.x}, y: {self.y}, h: {self.z}"
    
    def __str__(self):
        return f"x: {self.x}, y: {self.y}, h: {self.z}"


#Direct geodetic problem
def direct_geo_task(x=0, y=0, d=0, a=0):
    return {"x" : x + d * cos(a), "y" : y + d * sin(a)}


#Invert geodetic problem
def inverse_geo_task(x1=0, y1=0, x2=0, y2=0):
    dx, dy = x2 - x1, y2 - y1
    d = sqrt(pow(dx,2)+pow(dy,2))
    r = atan(abs(dy / dx))
    
    if dx > 0 and dy > 0:
        a = r
    elif dx < 0 and dy > 0:
        a = 180 - r
    elif dx < 0 and dy < 0:
        a = 180 + r
    elif dx > 0 and dy < 0:
        a = 360 - r
    elif dx == 0:
        if dy > 0:
            a = 90
        elif dy < 0:
            a = 270
    elif dy == 0:
        if dx > 0:
            a = 0
        elif dx < 0:
            a = 180
    else:
        a = None
        
    return {"d" : d, "a" : a}


#Horizontal distance
def horizontal_distance(s=0, v=0):
    return s * cos(v)


from pydesy.gmath import *

def perpendicular_baseline(s, basis, b):
    if basis < 0:
        raise Exception("The basis with cannot be negative or equal to zero")
    g = asin(s / basis * sin(b))
    x = s * cos(180 - (b + g))
    y = s * sin(180 - (b + g)) * cot(g)
    d = y * tan(g)
    
    return {"x" : x, "y" : y, "d" : d, "g" : g}


import math

r = 180 / math.pi * 3600

def angle(d=0, m=0, s=0):
    return revs(d + m / 60 + s / 3600)

def angdms(degrees=0):
    degrees = revs(degrees)
    d = int(degrees)
    m = int((degrees - int(degrees)) * 60)
    s = ((degrees - int(degrees)) * 60 - m) * 60
    return {"d" : d, "m" : m, "s" : s}

def revs(degrees):
    while degrees > 360:
        degrees -= 360
    while degrees < 0:
        degrees += 360
    return degrees

def rad(a=1):
    return a * math.pi / 180

def deg(a=1):
    return a * 180 / math.pi

def dirang(direction, horizontalAngle):
    return revs(angle(direction + horizontalAngle) + (-180 if angle(direction + horizontalAngle) >= 180 else 180))

def change_angle(a):
    return revs(angle(360 - a))

def sin(a=0):    
    return math.sin(rad(a))

def cos(a=0):
    return math.cos(rad(a))

def tan(a=0):
    return math.tan(rad(a))

def cot(a=0):
    return 1 / math.tan(rad(a))

def asin(x):
    return deg(math.asin(x))

def acos(x):
    return deg(math.acos(x))

def atan(x):
    return deg(math.atan(x))

def acot(x):
    return deg(math.atan(1 / x))

def sqrt(x):
    return math.sqrt(x)


import matplotlib.pyplot as plt

from pydesy.basic import *
from pydesy.gmath import *

class PolarSerif:
    def __init__(self, A, B):
        self.A = A
        self.B = B
        self.directionAngle = A.igt(B)["a"]

    def set_side_lenghts(self, sideLenghts):
        self.sideLenghts = sideLenghts

    def set_horizontal_angles(self, horizontalAngles):
        self.horizontalAngles = horizontalAngles

    def set_points(self, points):
        self.points = points

    def calculate_serif(self):
        self.points = []
        for i in range(len(self.sideLenghts)):
            directionAngle = revs(self.directionAngle + self.horizontalAngles[i])
            print(directionAngle)
            dx = self.sideLenghts[i] * cos(directionAngle)
            dy = self.sideLenghts[i] * sin(directionAngle)
            self.points.append(Point(self.A.x + dx, self.A.y + dy))
        return self.points

    def calculate_accuracy_score(self, lenghtError, angleError):
        self.errors = []
        for i in range(len(self.points)):
            self.errors.append(self.sideLenghts[i] ** 2 * angleError ** 2 / r ** 2 + lenghtError ** 2)
        return self.errors

    def draw_serif(self):
        plt.figure("Polar serif")
        plt.text(self.A.y, self.A.x, "A")
        plt.text(self.B.y, self.B.x, "B")
        for i, point in enumerate(self.points):
            plt.plot([self.A.y, point.y], [self.A.x, point.x], linestyle="--", color="red", marker=".")
            plt.text(point.y, point.x, f"{i}")

        plt.plot([self.A.y, self.B.y], [self.A.x, self.B.x], color="blue", marker="^")
        plt.xlabel("y")
        plt.ylabel("x")
        plt.title("Polar serif")
        plt.show()


from pydesy.gmath import *
from pydesy.basic import *
def polygon_area(polygon):
    countPolygons = len(polygon)    
    if countPolygons < 3:
        raise Exception("To solve this problem, you need at least 3 points")
        
    firstSum = sum([polygon[i].x * polygon[i+1].y for i in range(countPolygons - 1)]) +  polygon[countPolygons - 1].x * polygon[0].y
    secondSum = sum([polygon[i+1].x * polygon[i].y for i in range(countPolygons - 1)]) +  polygon[countPolygons - 1].y * polygon[0].x
    return (firstSum - secondSum) / 2

def three_sides(a, b, c):
    A = acos(b ^ 2 + c ^ 2 - a ^ 2 / (2 * b * c))
    B = acos(a ^ 2 + c ^ 2 - b ^ 2 / (2 * a * c))
    C = acos(a ^ 2 + b ^ 2 - c ^ 2 / (2 * a * b))

    return {
        "A" : A,
        "B" : B,
        "C" : C
    }


from setuptools import setup

setup(name='pydesy',
      version='0.0.2',
      description='Geodetic library',
      packages=['pydesy'],
      author_email='shmakov.maxim2001@yandex.ru',
      zip_safe=False)
